/***********************************************************************
 * File: days.c
 * 	(Extracted from
 *	   "$Id: days.c,v 1.15 2007/04/29 13:42:59 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C source implements "is_valid_date()" and "date_in_English()"
 * API-s.
 *
 * Copyright (C) 2005-2007 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#include "days.h"

/* get the year, month and day from a numeric string. */
static int _get_ymd(char	*date
		, size_t	sz
		, unsigned long	*year
		, int		*month
		, int		*day
		)
{
	char	sv, *endp = NULL;

	/* test: the arguments are illegal? */
	if ((((((date == NULL)
	|| sz < 5)
	|| !isdigit(date[sz - 1]))
	|| !isdigit(date[sz - 2]))
	|| !isdigit(date[sz - 3]))
	|| !isdigit(date[sz - 4])) {
		return	DAYS_ILLEGAL_ARGS;
	}

	/* set the month and day pointing areas */
	*month	= (10 * (int)(date[sz - 4] - '0'))
				+ (int)(date[sz - 3] - '0');
	*day	= (10 * (int)(date[sz - 2] - '0'))
				+ (int)(date[sz - 1] - '0');

	/* test: the month is illegal? */
	if ((((*month < 1) || *month > 12)
	|| *day < 1) || *day > 31) {
		return	DAYS_ILLEGAL_ARGS;
	}

	/* convert the year part of the date string
	 * into the unsigned long
	 */
	sv	= date[sz - 4];
	date[sz - 4]	= '\0';
	errno	= 0; /* initialize errno */
	*year	= strtoul((const char *)date, &endp, 10);
	date[sz - 4]	= sv;

	/* test: the year part is over flow? */
	if ((*year == ULONG_MAX) && errno == ERANGE) {
		return	DAYS_ILLEGAL_ARGS;
	}

	/* test: the size of date string is least than 8 (yyyymmdd)? */
	if (sz < 8) {
		/* amend as the 21st century */
		*year	+= DAYS_BASE_YEAR;
	}

	return	(*year < 1)? DAYS_ILLEGAL_ARGS: 0;
}

/* validate the result of `_get_ymd()' (year, month and day) */
static int	days_a_month[] = { 31, 28, 31, 30, 31, 30
				 , 31, 31, 30, 31, 30, 31, 29
				 };

static int _validate_ymd(unsigned long	year
			, int		month
			, int		day
			)
{
	int	result;

	/* test: the tuple of year, month and day is valid? */
	if (month == 1 || month > 2) {
		result	= (day <= days_a_month[month - 1])?
					DAYS_TRUE: DAYS_FALSE;
	}
	else if (IS_LEAP_YEAR(year)) {
		result	= (day <= days_a_month[12])?
					DAYS_TRUE: DAYS_FALSE;
	}
	else {
		result	= (day <= days_a_month[1])?
					DAYS_TRUE: DAYS_FALSE;
	}

	return	result; /* DAYS_TRUE or DAYS_FALSE */
}

/* This determines the numeric string means a valid date. */
static int _is_valid_date(char *date
			, size_t sz
			, unsigned long *year
			, int *month
			, int *day
			)
{
	/* get the year, month and day from date string */
	if (_get_ymd(date, sz, year, month, day) != 0) {
		return	DAYS_ILLEGAL_ARGS;
	}

	/* test: the tuple of year, month and day is valid? */
	return	_validate_ymd(*year, *month, *day);
				/* DAYS_TRUE:  valid date string
				 * DAYS_FALSE: invalid date string
				 */
}


/*
 * This API determines the numeric string is valid date (yyy...yyymmdd)
 *
 * arguments:
 *	date is pointing to the date string (yyy...yyymmdd).
 *	sz indicates the size of str pointing area.
 *
 * returns:
 *	DAYS_ILLEGAL_ARGS (= -1) : illegal arguments
 *	DAYS_FALSE (= 0)         : the string doesn't means valid date.
 *	DAYS_TRUE  (= 1)         : the string means valid date.
 */
int is_valid_date(char *date, size_t sz)
{
	unsigned long	year = 0UL;
	int		month = 0, day = 0;

	return	_is_valid_date(date, sz, &year, &month, &day);
}


/*
 * This API gets the year, month and day from a date string.
 *
 * arguments:
 *	 date points to the numeric string.
 *	 sz indicates the length of date strings.
 *	 year points to the integer as year.
 *	 month points to the integer as month.
 *	 day points to the integer as day.
 *
 * returns:
 *	DAYS_ILLEGAL_ARGS : it means illegal arguments.
 *	0 : normal end
 */
int get_ymd(char	*date
	, size_t	sz
	, unsigned long	*year
	, int		*month
	, int		*day
	)
{
	if (((((year == NULL) || month == NULL) || day == NULL)
	|| _get_ymd(date, sz, year, month, day) != 0)
	|| _validate_ymd(*year, *month, *day) != DAYS_TRUE) {
		return	DAYS_ILLEGAL_ARGS;
	}

	return	0; /* normal end */
}


/*
 * This API translate the numeric string into date in English.
 *
 * arguments:
 *	date is pointing to the numeric string (yyy...yyymmdd).
 *	sz indicates the size of str pointing area.
 *	fmt indicates the format of European or American.
 *	str is pointing the output string.
 *	str_sz indicates the size of the str output area.
 *
 * returns:
 *	NULL : illegal arguments
 *	non-NULL : the translated string address (same as str)
 */
char *date_in_English(char *date
		, size_t sz
		, DAYS_FMT fmt
		, char *str
		, size_t str_sz
		)
{
	unsigned long	year = 0UL;
	int		month = 0, day = 0;
	char		*mon_names[] = {
				  "January", "February", "March"
				, "April",   "May",      "June"
				, "July",    "August",   "September"
				, "October", "November", "December"
				}
			, *fmtstr;

	/* test: arguments are illegal? */
	if (((((str == NULL)
	|| fmt < DAYS_FMT_EUROPEAN)
	|| fmt > DAYS_FMT_AMERICAN)
	|| _is_valid_date(date, sz, &year, &month, &day) != DAYS_TRUE)
	|| str_sz < ((sz < 8)? 8: sz) + strlen(mon_names[month - 1])
					+ ((day >= 10)? 2: 1)) {
		return	NULL;
	}

	/* which format? */
	if (fmt == DAYS_FMT_EUROPEAN) {
		fmtstr	= ((day >= 10)? "%2d, %s %lu": "%1d, %s, %lu");
		(void)snprintf(str, str_sz
			, (const char *)fmtstr
			, day, mon_names[month - 1], year
			);
	}
	else {	/* in American English */
		fmtstr	= ((day >= 10)? "%s, %2d %lu": "%s,%2d, %lu");
		(void)snprintf(str, str_sz
			, (const char *)fmtstr
			, mon_names[month - 1], day, year
			);
	}

	return	str; /* translated date string */
}


/*
 * This API returns the day of year.
 *
 * arguments:
 *	date is pointing to the date string.
 *	sz indicates the length of the date string.
 *
 * returns:
 *	zero : illegal arguments
 *	positive integers: the Julian date for date string
 */
/* table of the past days of the month */
static int	past_days[] =		{   0,  31,  59,  90
					, 120, 151, 181, 212
					, 243, 273, 304, 334
					, 365
					}
		, leap_past_days[] =	{   0,  31,  60,  91
					, 121, 152, 182, 213
					, 244, 274, 305, 335
					, 366
					};

/* This returns the day of year for year/month/day. */
static int _get_day_of_year(unsigned long year, int month, int day)
{
	return	past_days[month - 1]
		+ day
		+ (((month > 2) && IS_LEAP_YEAR(year))? 1: 0)
		;
}

int get_day_of_year(char *date, size_t sz)
{
	unsigned long	year = 0UL;
	int		month = 0, day = 0;

	/* test: arguments are illegal? */
	if (((date == NULL)
	|| sz == 0)
	|| _is_valid_date(date, sz, &year, &month, &day) != DAYS_TRUE) {
		return	0;
	}

	return	_get_day_of_year(year, month, day);
}


/*
 * This API counts the days since 1, January AD.1 (it's the 1st day).
 *
 * arguments:
 *	date is pointing to the date.
 *	sz indicates the length of the date string.
 *
 * returns:
 *	LLONG_MIN :		illegal arguments
 *	0LL to 1568704592244LL:	the count of days.
 */
static long long _days_count(unsigned long year, int month, int day)
{
	long long	result = (long long)(year - 1);
	result	= result * 365LL + result / 400LL
				- result / 100LL + result / 4LL;
	return	result	+ _get_day_of_year(year, month, day);
}

long long days_count(char *date, size_t sz)
{
	unsigned long	year = 0UL;
	int		month = 0, day = 0;

	if ((date == NULL)
	|| _is_valid_date(date, sz, &year, &month, &day) != DAYS_TRUE) {
		return	LLONG_MIN; /* Failed: illegal arguments */
	}

	return	_days_count(year, month, day);
}


/*
 * This API returns pointer to the name of the day of week.
 *
 * arguments:
 *	days_counts is returns of `days_count()' API.
 *
 * returns:
 *	NULL	: illegal arguments
 *	non-NULL: address of the name of the day of week.
 */
char *day_of_week_str(long long days_counts)
{
	static char *day_of_week_strs[] = {
		  "Sunday",	"Monday",	"Tuesday"
		, "Wednesday",	"Thursday",	"Friday", "Saturday"
		};
	Day_of_week_e	dow = day_of_week(days_counts);
	char		*result;

	result	= ((dow != Illegal_day_of_week)?
				/* returns address of name */
				day_of_week_strs[dow]
			:	NULL	/* illegal arguments */
			);
	return	result;
}


/*
 * This API subtracts the day-counts of date1 and date2.
 *
 * arguments:
 *	date1 is pointing to the subtractive date string.
 *	sz1 indicates the length of date1 staring.
 *	date2 is pointing to the subtracter date string.
 *	sz2 indicates the length of date2 staring.
 *
 * returns:
 *	LLONG_MIN : illegal arguments
 *	-1568704592244LL to 1568704592244LL: the diff of day-counts.
 */
long long diff_dates(char *date1, size_t sz1, char *date2, size_t sz2)
{
	long long	d1, d2 = LLONG_MIN;

	if (((d1 = days_count(date1, sz1)) == LLONG_MIN)
	|| (d2 = days_count(date2, sz2)) == LLONG_MIN) {
		return	LLONG_MIN;
	}

	return	d1 - d2;
}


/*
 * This API translate day-count to the 'YYY...YYYMMDD' formatted string.
 *
 * arguments:
 *	  day_count indicates the count of past days that returns of
 *	`days_count()' API.
 *	  date points to the buffer that is outputted 'YYY...YYYMMDD'
 *	formatted string.
 *	  sz indicates the size of buffer that is pointed from `date'.
 *	sz must be greater than nine bytes.
 *
 * returns:
 *	NULL: invalid arguments
 *	non-NULL: normal end (same as `date')
 */
char *days_count_to_date_str(long long day_count, char *date, size_t sz)
{
	int		*past;
	char		*fmt;
	size_t		m = 0;
	long long	year = 0, d = day_count;

	/* test: the arguments are illegal? */
	if (((day_count <= 0LL) || date == NULL) || sz < 9) {
		errno	= EINVAL;
		return NULL; /* ERROR: invalid arguments */
	}

	/* calculate `d' */
	do {
		year	+= (d / 366);
		d = day_count - (365 * year)
			- (year / 4 - year / 100 + year / 400);
	} while (d > ((IS_LEAP_YEAR(year + 1))? 366: 365));

	/* now, `year' indicates the last year */
	++year; /* increment `year' */

	/* determine month `m'... */
	for (past = (IS_LEAP_YEAR(year))? leap_past_days: past_days
		, m = 1
	; (m <= 12) && d > past[m]
	; ++m) { ; }

	d	-= past[m - 1]; /* determine day `d' */

	/* if year indicates between the 1st and 10th centuries, we must
	 * format it to the zero filled 4 digits.  otherwise, we must
	 * format it as is.
	 */
	fmt	= ((year < 1000)?	"%04lld%02d%02lld"
			:		"%lld%02d%02lld"
			);
	(void)snprintf(date, sz, (const char *)fmt, year, m, d);


	return	date; /* normal end (address of output buffer `date') */
}


/*
 * This API calculate durations for years, months, weeks.
 *
 * arguments:
 *	 date points to a numeral string.
 *	 unit indicates the duration unit (Duration_unit_e).
 *	 delta indicates the delta for date argument.
 *	 duration points to the result buffer of calculating duration.
 *
 * returns:
 *	NULL: invalid arguments
 *	non-NULL: same as duration argument
 */
Duration_t *calc_duration(char			*date
			, Duration_unit_e	unit
			, long long		delta
			, Duration_t		*duration
			)
{
	Duration_t	*result = NULL;
	unsigned long	year = 0UL;
	int		month = 0, day = 0;

	if (((date == NULL) || duration == NULL)
	|| _is_valid_date(date, strlen(date), &year, &month, &day)
							!= DAYS_TRUE) {
		errno	= EINVAL;
		return	result;	/* ERROR: invalid arguments */
	}

	switch (unit) {
	case Duration_unit_Years:
		if ((delta < 0UL && llabs(delta) >= (long long)year)
		|| (delta > 0UL
			&& delta > (long long)(ULONG_MAX - year))
		) {
			errno	= EINVAL; /* ERROR: invalid arguments */
		}
		else {
			result	= duration;
			year	+= delta;
			result->begin_of = _days_count(year, 1, 1);
			result->end_of   = _days_count(year, 12, 31);
		}
		break;

	case Duration_unit_Months: {
		long long this_month = ((long long)year - 1LL) * 12LL
						+ (long long)month;
		if ((delta < 0UL
			&& llabs(delta) >= this_month)
		|| (delta > 0UL
			&& delta > (long long)(ULONG_MAX * 12
							- this_month))
		) {
			errno	= EINVAL; /* ERROR: invalid arguments */
		}
		else {
			result		= duration;
			this_month	+= delta;
			year		= (unsigned long)(
						(this_month + 11) / 12);
			month		= (int)((this_month - 1) % 12);
			if (IS_LEAP_YEAR(year) && month == 1) {
				day	= days_a_month[
					  	(sizeof(days_a_month) /
						sizeof(days_a_month[0]))
						- 1
						];
				month	= 2;
			}
			else {
				day	= days_a_month[month];
				++month;
			}
			result->begin_of = _days_count(year, month, 1);
			result->end_of  = _days_count(year, month, day);
		}
		} break;

	case Duration_unit_Weeks: {
		long long this_week = _days_count(year, month, day);
		this_week	-= (long long)(day_of_week(this_week));
		if (((this_week < 0)
		|| (delta < 0 && llabs(delta) * 7 >= this_week))
		|| (delta > 0 && delta * 7 > MAX_DAYS_COUNT - this_week)
		) {
			errno	= EINVAL; /* ERROR: invalid arguments */
		}
		else {
			result	= duration;
			duration->begin_of = this_week + (delta * 7);
			duration->end_of   = duration->begin_of + 6;
		}
		} break;

	default:
		errno	= EINVAL;	/* ERROR: invalid arguments */
		break;
	}

	if (result != NULL) {
		result->days = (result->end_of - result->begin_of + 1);
	}

	return	result;
}
